#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include <vector>
#include <cstdint>
#include <cstdlib>
#include <cmath>
#include <cstring>

constexpr int SCREEN_WIDTH = 396;
constexpr int SCREEN_HEIGHT = 224;
constexpr int NUM_STARS = 10000;                     // 3200 stars can be easily managed by the fxCG 50
constexpr int FIXED_SHIFT = 16;
constexpr int FIXED_ONE = 1 << FIXED_SHIFT;
constexpr int32_t PI_FIXED = static_cast<int32_t>(3.1415926535f * FIXED_ONE);
constexpr int SIN_TABLE_SIZE = 360;

int32_t fixedMul(int32_t a, int32_t b) {
    return (static_cast<int64_t>(a) * b) >> FIXED_SHIFT;
}

int32_t fixedDiv(int32_t a, int32_t b) {
    return static_cast<int32_t>((static_cast<int64_t>(a) << FIXED_SHIFT) / b);
}

int32_t floatToFixed(float f) {
    return static_cast<int32_t>(f * FIXED_ONE);
}

int fixedToInt(int32_t f) {
    return f >> FIXED_SHIFT;
}

int sinTable[SIN_TABLE_SIZE];
int cosTable[SIN_TABLE_SIZE];

void initTrigTables() {
    for (int i = 0; i < SIN_TABLE_SIZE; ++i) {
        float angle = i * 3.1415926535f / 180.0f;
        sinTable[i] = floatToFixed(std::sin(angle));
        cosTable[i] = floatToFixed(std::cos(angle));
    }
}

int32_t fixedSin(int degrees) {
    return sinTable[degrees % SIN_TABLE_SIZE];
}

int32_t fixedCos(int degrees) {
    return cosTable[degrees % SIN_TABLE_SIZE];
}

struct Vec3Fix {
    int32_t x, y, z;
};

struct StarFix {
    Vec3Fix position;
    SDL_Color color;
};

Vec3Fix rotateX(const Vec3Fix& p, int angle) {
    int32_t sinA = fixedSin(angle);
    int32_t cosA = fixedCos(angle);
    return {
        p.x,
        fixedMul(p.y, cosA) - fixedMul(p.z, sinA),
        fixedMul(p.y, sinA) + fixedMul(p.z, cosA)
    };
}

Vec3Fix rotateY(const Vec3Fix& p, int angle) {
    int32_t sinA = fixedSin(angle);
    int32_t cosA = fixedCos(angle);
    return {
        fixedMul(p.x, cosA) - fixedMul(p.z, sinA),
        p.y,
        fixedMul(p.x, sinA) + fixedMul(p.z, cosA)
    };
}

Vec3Fix rotateZ(const Vec3Fix& p, int angle) {
    int32_t sinA = fixedSin(angle);
    int32_t cosA = fixedCos(angle);
    return {
        fixedMul(p.x, cosA) - fixedMul(p.y, sinA),
        fixedMul(p.x, sinA) + fixedMul(p.y, cosA),
        p.z
    };
}

SDL_Color colorFromTemperatureFixed(int32_t tempK) {
    int32_t t = fixedDiv(tempK - floatToFixed(3000.0f), floatToFixed(7000.0f));
    int32_t t_abs = std::abs(t - (FIXED_ONE >> 1));

    int32_t r = std::min(255, fixedToInt(fixedMul(FIXED_ONE * 255, std::min(FIXED_ONE, fixedMul(floatToFixed(1.5f), t)))));
    int32_t g = std::min(255, fixedToInt(floatToFixed(200.0f) - fixedMul(floatToFixed(100.0f), t_abs)));
    int32_t b = std::max(0, fixedToInt(fixedMul(FIXED_ONE * 255, FIXED_ONE - t)));

    return {static_cast<Uint8>(r), static_cast<Uint8>(g), static_cast<Uint8>(b), 255};
}

std::vector<StarFix> generateStars() {
    std::vector<StarFix> stars;
    const int numArms = 4;
    const int32_t twistFactor = floatToFixed(5.0f);
    const int32_t minRadius = floatToFixed(10.0f);
    const int32_t maxRadius = floatToFixed(170.0f);

    for (int i = 0; i < NUM_STARS; ++i) {
        float t = static_cast<float>(rand()) / RAND_MAX;
        float sqrtT = std::sqrt(t);
        int32_t ft = floatToFixed(t);
        int32_t radius = minRadius + fixedMul((FIXED_ONE - floatToFixed(sqrtT)), (maxRadius - minRadius));

        int arm = i % numArms;
        int32_t armOffset = floatToFixed((2.0f * 3.1415926535f / numArms) * arm);
        int32_t angle = fixedMul(twistFactor, fixedDiv(radius, maxRadius)) + armOffset;

        float jitter = ((rand() % 1000) / 750.0f - 0.5f) * 3.1415926535f / 10.0f;
        angle += floatToFixed(jitter);

        float radiusJitter = ((rand() % 1000) / 500.0f - 0.5f) * 10.0f;
        radius += floatToFixed(radiusJitter);

        int deg = (fixedToInt(fixedMul(angle, floatToFixed(180.0f / 3.1415926535f)))) % 360;

        int32_t x = fixedMul(radius, fixedCos(deg));
        int32_t y = fixedMul(radius, fixedSin(deg));
        int32_t z = floatToFixed(((rand() % 1000) / 1000.0f - 0.5f) * 10.0f);

        float rTemp = static_cast<float>(rand()) / RAND_MAX;
        float tempK = 3000.0f + std::pow(1.0f - rTemp, 2.0f) * 7000.0f;
        int32_t tempKFix = floatToFixed(tempK);

        SDL_Color c = colorFromTemperatureFixed(tempKFix);
        stars.push_back({{x, y, z}, c});
    }

    return stars;
}

void drawStar(SDL_Renderer* renderer, int x, int y, SDL_Color color, int size) {
    SDL_SetRenderDrawColor(renderer, color.r, color.g, color.b, color.a);
    for (int dx = -size; dx <= size; ++dx) {
        for (int dy = -size; dy <= size; ++dy) {
            if (dx * dx + dy * dy <= size * size)
                SDL_RenderDrawPoint(renderer, x + dx, y + dy);
        }
    }
}

int main() {
    SDL_Init(SDL_INIT_VIDEO);
    IMG_Init(IMG_INIT_PNG);

    initTrigTables();

    SDL_Window* window = SDL_CreateWindow("Galaxie", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
                                          SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_SHOWN);
    SDL_Renderer* renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED);

    SDL_Surface* imageSurface = IMG_Load("shmup.png");
    if (!imageSurface) {
        SDL_Log("Erreur chargement image : %s", IMG_GetError());
        return 1;
    }
    SDL_Texture* imageTexture = SDL_CreateTextureFromSurface(renderer, imageSurface);
    int imgW = imageSurface->w;
    int imgH = imageSurface->h;
    SDL_FreeSurface(imageSurface);



    std::vector<StarFix> stars = generateStars();

    int cameraYaw = 345;
    int cameraPitch = 300;
    int galaxyRotation = 0;
    int32_t focalLength = floatToFixed(1000.0f);

    bool running = true;
    while (running) {
        SDL_Event e;
        while (SDL_PollEvent(&e)) {
            if (e.type == SDL_QUIT) running = false;
            else if (e.type == SDL_KEYDOWN) {
                switch (e.key.keysym.sym) {
                    case SDLK_LEFT: cameraYaw -= 5; break;
                    case SDLK_RIGHT: cameraYaw += 5; break;
                    case SDLK_UP: cameraPitch -= 5; break;
                    case SDLK_DOWN: cameraPitch += 5; break;
                }
            }
        }

        galaxyRotation -= 1;
        if (galaxyRotation < 0)  galaxyRotation += 360;

        SDL_SetRenderDrawColor(renderer, 0, 0, 16, 255);
        SDL_RenderClear(renderer);

        for (const StarFix& s : stars) {
            Vec3Fix p = rotateZ(s.position, galaxyRotation);
            p = rotateY(p, cameraYaw);
            p = rotateX(p, cameraPitch);

            int32_t denom = focalLength + p.z;
            if (denom <= 0) continue;

            int32_t scale = fixedDiv(focalLength, denom);
            int screenX = (fixedMul(p.x, scale) >> FIXED_SHIFT) + SCREEN_WIDTH / 2;
            int screenY = (fixedMul(p.y, scale) >> FIXED_SHIFT) + SCREEN_HEIGHT / 2;

            int size = std::max(1, (scale * 2) >> FIXED_SHIFT);
            drawStar(renderer, screenX, screenY, s.color, size);
        }

        SDL_Rect dstRect;
        dstRect.x = (SCREEN_WIDTH - imgW) / 2;
        dstRect.y = (SCREEN_HEIGHT - imgH) / 4;
        dstRect.w = imgW;
        dstRect.h = imgH;
        SDL_RenderCopy(renderer, imageTexture, nullptr, &dstRect);

        SDL_RenderPresent(renderer);
        SDL_Delay(16);
    }

    SDL_DestroyRenderer(renderer);
    SDL_DestroyTexture(imageTexture);
    SDL_DestroyWindow(window);
    IMG_Quit();
    SDL_Quit();
    return 0;
}
